==> ThbgmExtractor-1.6.5.tar.bz2: ThbgmExtractor-1.6.5/core/Smdn.Windows.Forms.ThbgmPlayer/Smdn.Windows.Forms.ThbgmPlayer/Player.cs <==
// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2009-2019 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Drawing;
using System.IO;
using System.Threading;
using System.Windows.Forms;

using Smdn.Formats.Thbgm;
using Smdn.Media;

namespace Smdn.Windows.Forms.ThbgmPlayer {
  public partial class Player : UserControl {
    /*
     * class members
     */
    public const int MaxVolume = 100;
    public const int MinVolume = 0;
    public const int DefaultVolume = 75;
    public const int DefaultWaveFormZoomScale = WaveFormView.ZoomScaleDefault;

    static Player()
    {
      imageButtonPlay = new Bitmap(24, 24);
      imageButtonStop = new Bitmap(24, 24);
      imageButtonPause = new Bitmap(24, 24);
      imageButtonPrev = new Bitmap(24, 24);
      imageButtonNext = new Bitmap(24, 24);

      var images = new[] {
        imageButtonPlay,
        imageButtonStop,
        imageButtonPause,
        imageButtonPrev,
        imageButtonNext,
      };

      for (var i = 0; i < images.Length; i++) {
        var image = images[i];

        using (var g = Graphics.FromImage(image)) {
          g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;
          g.Clear(Color.Transparent);

          switch (i) {
            case 0: { // play
              var trianglePoints = new[] {
                new PointF(image.Width * 0.2f, image.Height * 0.2f),
                new PointF(image.Width * 0.2f, image.Height * 0.8f),
                new PointF(image.Width * 0.8f, image.Height * 0.5f),
              };
              g.FillPolygon(SystemBrushes.ControlText, trianglePoints);
              break;
            }

            case 1: { // stop
              var rectangle = RectangleF.FromLTRB(image.Width * 0.2f, image.Height * 0.2f, image.Width * 0.8f, image.Height * 0.8f);
              g.FillRectangle(SystemBrushes.ControlText, rectangle);
              break;
            }

            case 2: { // pause
              var leftRectangle  = RectangleF.FromLTRB(image.Width * 0.2f, image.Height * 0.2f, image.Width * 0.4f, image.Height * 0.8f);
              var rightRectangle = RectangleF.FromLTRB(image.Width * 0.6f, image.Height * 0.2f, image.Width * 0.8f, image.Height * 0.8f);
              g.FillRectangle(SystemBrushes.ControlText, leftRectangle);
              g.FillRectangle(SystemBrushes.ControlText, rightRectangle);
              break;
            }

            case 3: { // prev
              g.FillRectangle(SystemBrushes.ControlText, RectangleF.FromLTRB(image.Width * 0.1f, image.Height * 0.2f, image.Width * 0.2f, image.Height * 0.8f));
              var trianglePoints1 = new[] {
                new PointF(image.Width * 0.2f, image.Height * 0.5f),
                new PointF(image.Width * 0.5f, image.Height * 0.8f),
                new PointF(image.Width * 0.5f, image.Height * 0.2f),
              };
              g.FillPolygon(SystemBrushes.ControlText, trianglePoints1);
              var trianglePoints2 = new[] {
                new PointF(image.Width * 0.5f, image.Height * 0.5f),
                new PointF(image.Width * 0.8f, image.Height * 0.8f),
                new PointF(image.Width * 0.8f, image.Height * 0.2f),
              };
              g.FillPolygon(SystemBrushes.ControlText, trianglePoints2);
              break;
            }

            case 4: { // next
              using (var flipped = (Bitmap)imageButtonPrev.Clone()) {
                flipped.RotateFlip(RotateFlipType.RotateNoneFlipX);
                g.DrawImageUnscaled(flipped, 0, 0);
              }
              break;
            }
          }
        }
      }
    }

    private static Bitmap imageButtonPlay;
    private static Bitmap imageButtonStop;
    private static Bitmap imageButtonPause;
    private static Bitmap imageButtonPrev;
    private static Bitmap imageButtonNext;

    /*
     * events
     */
    public event EventHandler PlayDone;
    public event EventHandler PlayerStatusChanged;
    public event EventHandler TrackChanged;
    public event EventHandler PrevButtonClick;
    public event EventHandler NextButtonClick;

    /*
     * properties
     */
    public TimeSpan Position {
      get {
        RejectDisposed();
        if (player == null)
          return TimeSpan.MinValue;
        else
          return player.Position;
      }
      set {
        RejectDisposed();
        if (player != null)
          player.Position = value;
      }
    }

    public Track Track {
      get {
        RejectDisposed();
        return track;
      }
      set {
        RejectDisposed();
        OnTrackChanged(value);
      }
    }

    public string BgmSourcePath {
      get { return bgmSourcePath; }
      set {
        if (value == bgmSourcePath)
          return;

        bgmSourcePath = value;

        Stop();

        buttonPlay.Enabled = File.Exists(bgmSourcePath);

        waveFormView.BgmSourcePath = BgmSourcePath;
      }
    }

    public int Volume {
      get { return trackBarVolume.Value; }
      set { trackBarVolume.Value = value; }
    }

    public bool Repeat {
      get { return checkBoxRepeat.Checked; }
      set {
        RejectDisposed();

        checkBoxRepeat.Checked = value;

        if (player != null)
          player.Repeat = value;
      }
    }

    public bool WaveFormAutoScroll {
      get { return checkBoxAutoScroll.Checked; }
      set {
        RejectDisposed();

        checkBoxAutoScroll.Checked = value;

        waveFormView.SeekBarAutoTracking = value;
      }
    }

    public int WaveFormZoomScale {
      get { return waveFormView.ZoomScale; }
      set {
        RejectDisposed();

        waveFormView.ZoomScale = Math.Min(Math.Max(value, WaveFormView.ZoomScaleMin), WaveFormView.ZoomScaleMax);
      }
    }

    public bool WaveFormDisplayWaveForm {
      get => waveFormView.DisplayWaveForm;
      set {
        RejectDisposed();

        waveFormView.DisplayWaveForm = value;
      }
    }

    public bool IsPlaying {
      get { return (player == null) ? false : player.Playing; }
    }

    public bool IsPausing {
      get { return (player == null) ? false : player.Pausing; }
    }

    public bool ShowPrevNextButtons {
      get { return showPrevNextButtons; }
      set
      {
        if (showPrevNextButtons != value)
          SetPrevNextButtonVisibility(value);
      }
    }

    /*
     * construction/destruction
     */
    public Player()
    {
      InitializeComponent();

      player = PcmStreamPlayer.Create();
      player.PlayDone += Player_PlayDone;

      waveFormView.SeekBarMoved += delegate(object sender, SeekBarMovedEventArgs e) {
        if (player == null || !player.Playing)
          e.Cancel = true;
        else
          player.Position = e.SeekBarPosition;
      };

      trackBarVolume.Minimum = MinVolume;
      trackBarVolume.Maximum = MaxVolume;
      trackBarVolume.SmallChange = (MaxVolume - MinVolume) / 100;
      trackBarVolume.LargeChange = (MaxVolume - MinVolume) / 10;
      trackBarVolume.Value = DefaultVolume;
      trackBarVolume.ValueChanged += delegate {
        SetPlayerVolume();
      };
      trackBarVolume.Scroll += delegate {
        SetVolumeToolTip();
      };
      trackBarVolume.MouseHover += delegate {
        SetVolumeToolTip();
      };

      buttonPlay.Click += delegate {
        if (player == null || !player.Playing)
          Play();
        else if (player.Pausing)
          Resume();
        else
          Pause();
      };

      buttonPlay.Image = imageButtonPlay;
      buttonPlay.ImageAlign = ContentAlignment.MiddleCenter;

      buttonStop.Click += delegate { Stop(); };
      buttonStop.Image = imageButtonStop;
      buttonStop.ImageAlign = ContentAlignment.MiddleCenter;
      buttonStop.Enabled = false;

      buttonPrev.Click += delegate { OnPrevButtonClick(EventArgs.Empty); };
      buttonPrev.Image = imageButtonPrev;
      buttonPrev.ImageAlign = ContentAlignment.MiddleCenter;

      buttonNext.Click += delegate { OnNextButtonClick(EventArgs.Empty); };
      buttonNext.Image = imageButtonNext;
      buttonNext.ImageAlign = ContentAlignment.MiddleCenter;

      SetPrevNextButtonVisibility(showPrevNextButtons);

      checkBoxRepeat.CheckedChanged += delegate {
        if (player != null)
          player.Repeat = checkBoxRepeat.Checked;
      };

      checkBoxAutoScroll.CheckedChanged += delegate {
        waveFormView.SeekBarAutoTracking = checkBoxAutoScroll.Checked;
      };

      SetPlayerVolume();

      timer = new System.Timers.Timer(100);
      timer.Enabled = false;
      timer.AutoReset = true;
      timer.SynchronizingObject = this;
      timer.Elapsed += delegate { SetSeekBarPosition(); };

      labelPosition.Left = checkBoxAutoScroll.Right + 3; // ???: anchor not works
    }

    private void Player_PlayDone(object sender, EventArgs e)
    {
      DisableStopButton();

      OnPlayDone(EventArgs.Empty);
      OnPlayerStatusChanged(EventArgs.Empty);
    }

    protected override void OnLoad(EventArgs e)
    {
      timer.Start();

      base.OnLoad(e);
    }

    private void DisposeInternal(bool disposing)
    {
      if (disposing) {
        if (!disposed) {
          Stop();

          if (timer != null) {
            timer.Stop();
            timer.Dispose();
            timer = null;
          }

          if (player != null) {
            player.PlayDone -= Player_PlayDone;
            player.Close();
            player = null;
          }

          if (stream != null) {
            stream.Close();
            stream = null;
          }

          disposed = true;
        }
      }
    }

    private void SetPrevNextButtonVisibility(bool visible)
    {
      showPrevNextButtons = visible;

      buttonPrev.Visible = showPrevNextButtons;
      buttonNext.Visible = showPrevNextButtons;
    }

    private void SetVolumeToolTip()
    {
      toolTipVolume.SetToolTip(trackBarVolume, GetVolumeInPercent().ToString("P0"));
    }

    /*
     * public methods
     */
    public void Play()
    {
      RejectDisposed();

      if (track == null)
        return;

      buttonPlay.Image = imageButtonPause;
      buttonStop.Enabled = true;

      try {
        stream = track.GetStream(BgmSourcePath);
      }
      catch (FileNotFoundException) {
        throw;
      }
      catch (ArgumentException) {
        buttonPlay.Image = imageButtonPlay;
        buttonStop.Enabled = false;
        return;
      }

      SetPlayerVolume();

      player.PlayAsync(stream, stream.Format.GetWaveFormatEx(), checkBoxRepeat.Checked);

      OnPlayerStatusChanged(EventArgs.Empty);
    }

    public void Stop()
    {
      RejectDisposed();

      if (player == null)
        return;

      DisableStopButton();

      player.Stop();

      OnPlayerStatusChanged(EventArgs.Empty);
    }

    public void Pause()
    {
      RejectDisposed();

      if (player == null)
        return;

      buttonPlay.Image = imageButtonPlay;

      player.Pause();

      OnPlayerStatusChanged(EventArgs.Empty);
    }

    public void Resume()
    {
      RejectDisposed();

      if (player == null)
        return;

      buttonPlay.Image = imageButtonPause;

      player.Resume();

      OnPlayerStatusChanged(EventArgs.Empty);
    }

    /*
     * private methods
     */
    private double GetVolumeInPercent()
    {
      return (trackBarVolume.Value - trackBarVolume.Minimum) / (double)(trackBarVolume.Maximum - trackBarVolume.Minimum);
    }

    private void SetPlayerVolume()
    {
      if (player == null)
        return;

      player.Volume = (int)(100 * GetVolumeInPercent());
    }

    private void DisableStopButton()
    {
      if (InvokeRequired) {
        Invoke((MethodInvoker)DisableStopButton);
        return;
      }

      buttonPlay.Image = imageButtonPlay;
      buttonStop.Enabled = false;
    }

    private void SetSeekBarPosition()
    {
      if (player == null)
        return;

      var position = player.Position;

      labelPosition.Text = string.Format("{0}:{1:D2}.{2:D3}", (int)position.TotalMinutes, position.Seconds, position.Milliseconds);

      if (IsPlaying && !IsPausing)
        waveFormView.SeekBarPosition = position;
    }

    private void OnTrackChanged(Track newTrack)
    {
      if (track == newTrack)
        return;

      Stop();

      if (stream != null) {
        stream.Close();
        stream = null;
      }

      track = newTrack;

      waveFormView.Track = track;

      effectEditor.Effect = (track == null) ? null : track.Effect;

      OnTrackChanged(EventArgs.Empty);
    }

    protected virtual void OnTrackChanged(EventArgs e)
    {
      var ev = this.TrackChanged;

      if (ev != null)
        ev(this, e);
    }

    protected virtual void OnPlayDone(EventArgs e)
    {
      var ev = this.PlayDone;

      if (ev != null)
        ev(this, e);
    }

    protected virtual void OnPlayerStatusChanged(EventArgs e)
    {
      var ev = this.PlayerStatusChanged;

      if (ev != null)
        ev(this, e);
    }

    protected virtual void OnPrevButtonClick(EventArgs e)
    {
      var ev = this.PrevButtonClick;

      if (ev != null)
        ev(this, e);
    }

    protected virtual void OnNextButtonClick(EventArgs e)
    {
      var ev = this.NextButtonClick;

      if (ev != null)
        ev(this, e);
    }

    private void RejectDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().Name);
    }

    private bool disposed = false;
    private PcmStreamPlayer player;
    private bool showPrevNextButtons = true;
    private Track track = null;
    private string bgmSourcePath = null;
    private EffectAppliedBgmStream stream = null;
    private System.Timers.Timer timer = null;
  }
}

==> ThbgmExtractor-1.6.6.tar.bz2: ThbgmExtractor-1.6.6/core/Smdn.Windows.Forms.ThbgmPlayer/Smdn.Windows.Forms.ThbgmPlayer/Player.cs <==
// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2009-2019 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Drawing;
using System.IO;
using System.Threading;
using System.Windows.Forms;

using Smdn.Formats.Thbgm;
using Smdn.Media;

namespace Smdn.Windows.Forms.ThbgmPlayer {
  public partial class Player : UserControl {
    /*
     * class members
     */
    public const int MaxVolume = 100;
    public const int MinVolume = 0;
    public const int DefaultVolume = 75;
    public const int DefaultWaveFormZoomScale = WaveFormView.ZoomScaleDefault;

    static Player()
    {
      imageButtonPlay = new Bitmap(24, 24);
      imageButtonStop = new Bitmap(24, 24);
      imageButtonPause = new Bitmap(24, 24);
      imageButtonPrev = new Bitmap(24, 24);
      imageButtonNext = new Bitmap(24, 24);
      imageButtonCanNotPlay = new Bitmap(24, 24);

      var images = new[] {
        imageButtonPlay,
        imageButtonStop,
        imageButtonPause,
        imageButtonPrev,
        imageButtonNext,
        imageButtonCanNotPlay,
      };

      for (var i = 0; i < images.Length; i++) {
        var image = images[i];

        using (var g = Graphics.FromImage(image)) {
          g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;
          g.Clear(Color.Transparent);

          switch (i) {
            case 0: { // play
              var trianglePoints = new[] {
                new PointF(image.Width * 0.2f, image.Height * 0.2f),
                new PointF(image.Width * 0.2f, image.Height * 0.8f),
                new PointF(image.Width * 0.8f, image.Height * 0.5f),
              };
              g.FillPolygon(SystemBrushes.ControlText, trianglePoints);
              break;
            }

            case 1: { // stop
              var rectangle = RectangleF.FromLTRB(image.Width * 0.2f, image.Height * 0.2f, image.Width * 0.8f, image.Height * 0.8f);
              g.FillRectangle(SystemBrushes.ControlText, rectangle);
              break;
            }

            case 2: { // pause
              var leftRectangle  = RectangleF.FromLTRB(image.Width * 0.2f, image.Height * 0.2f, image.Width * 0.4f, image.Height * 0.8f);
              var rightRectangle = RectangleF.FromLTRB(image.Width * 0.6f, image.Height * 0.2f, image.Width * 0.8f, image.Height * 0.8f);
              g.FillRectangle(SystemBrushes.ControlText, leftRectangle);
              g.FillRectangle(SystemBrushes.ControlText, rightRectangle);
              break;
            }

            case 3: { // prev
              g.FillRectangle(SystemBrushes.ControlText, RectangleF.FromLTRB(image.Width * 0.1f, image.Height * 0.2f, image.Width * 0.2f, image.Height * 0.8f));
              var trianglePoints1 = new[] {
                new PointF(image.Width * 0.2f, image.Height * 0.5f),
                new PointF(image.Width * 0.5f, image.Height * 0.8f),
                new PointF(image.Width * 0.5f, image.Height * 0.2f),
              };
              g.FillPolygon(SystemBrushes.ControlText, trianglePoints1);
              var trianglePoints2 = new[] {
                new PointF(image.Width * 0.5f, image.Height * 0.5f),
                new PointF(image.Width * 0.8f, image.Height * 0.8f),
                new PointF(image.Width * 0.8f, image.Height * 0.2f),
              };
              g.FillPolygon(SystemBrushes.ControlText, trianglePoints2);
              break;
            }

            case 4: { // next
              using (var flipped = (Bitmap)imageButtonPrev.Clone()) {
                flipped.RotateFlip(RotateFlipType.RotateNoneFlipX);
                g.DrawImageUnscaled(flipped, 0, 0);
              }
              break;
            }

            case 5: { // can not play
              var crossBar1Points = new[] {
                new PointF(image.Width * 0.1f, image.Height * 0.2f),
                new PointF(image.Width * 0.2f, image.Height * 0.1f),
                new PointF(image.Width * 0.9f, image.Height * 0.8f),
                new PointF(image.Width * 0.8f, image.Height * 0.9f),
              };
              g.FillPolygon(SystemBrushes.GrayText, crossBar1Points);

              var crossBar2Points = new[] {
                new PointF(image.Width * 0.9f, image.Height * 0.2f),
                new PointF(image.Width * 0.8f, image.Height * 0.1f),
                new PointF(image.Width * 0.1f, image.Height * 0.8f),
                new PointF(image.Width * 0.2f, image.Height * 0.9f),
              };
              g.FillPolygon(SystemBrushes.GrayText, crossBar2Points);
              break;
            }
          }
        }
      }
    }

    private static Bitmap imageButtonPlay;
    private static Bitmap imageButtonStop;
    private static Bitmap imageButtonPause;
    private static Bitmap imageButtonPrev;
    private static Bitmap imageButtonNext;
    private static Bitmap imageButtonCanNotPlay;

    /*
     * events
     */
    public event EventHandler PlayDone;
    public event EventHandler PlayerStatusChanged;
    public event EventHandler TrackChanged;
    public event EventHandler PrevButtonClick;
    public event EventHandler NextButtonClick;

    /*
     * properties
     */
    public TimeSpan Position {
      get {
        RejectDisposed();
        if (player == null)
          return TimeSpan.MinValue;
        else
          return player.Position;
      }
      set {
        RejectDisposed();
        if (player != null)
          player.Position = value;
      }
    }

    public Track Track {
      get {
        RejectDisposed();
        return track;
      }
      set {
        RejectDisposed();
        OnTrackChanged(value);
      }
    }

    public string BgmSourcePath {
      get { return bgmSourcePath; }
      set {
        if (value == bgmSourcePath)
          return;

        bgmSourcePath = value;

        Stop();

        buttonPlay.Enabled = File.Exists(bgmSourcePath);

        waveFormView.BgmSourcePath = BgmSourcePath;
      }
    }

    public int Volume {
      get { return trackBarVolume.Value; }
      set { trackBarVolume.Value = value; }
    }

    public bool Repeat {
      get { return checkBoxRepeat.Checked; }
      set {
        RejectDisposed();

        checkBoxRepeat.Checked = value;

        if (player != null)
          player.Repeat = value;
      }
    }

    public bool WaveFormAutoScroll {
      get { return checkBoxAutoScroll.Checked; }
      set {
        RejectDisposed();

        checkBoxAutoScroll.Checked = value;

        waveFormView.SeekBarAutoTracking = value;
      }
    }

    public int WaveFormZoomScale {
      get { return waveFormView.ZoomScale; }
      set {
        RejectDisposed();

        waveFormView.ZoomScale = Math.Min(Math.Max(value, WaveFormView.ZoomScaleMin), WaveFormView.ZoomScaleMax);
      }
    }

    public bool WaveFormDisplayWaveForm {
      get => waveFormView.DisplayWaveForm;
      set {
        RejectDisposed();

        waveFormView.DisplayWaveForm = value;
      }
    }

    public bool IsPlaying {
      get { return (player == null) ? false : player.Playing; }
    }

    public bool IsPausing {
      get { return (player == null) ? false : player.Pausing; }
    }

    public bool ShowPrevNextButtons {
      get { return showPrevNextButtons; }
      set
      {
        if (showPrevNextButtons != value)
          SetPrevNextButtonVisibility(value);
      }
    }

    /*
     * construction/destruction
     */
    public Player()
    {
      InitializeComponent();

      waveFormView.SeekBarMoved += delegate(object sender, SeekBarMovedEventArgs e) {
        if (player == null || !player.Playing)
          e.Cancel = true;
        else
          player.Position = e.SeekBarPosition;
      };

      trackBarVolume.Minimum = MinVolume;
      trackBarVolume.Maximum = MaxVolume;
      trackBarVolume.SmallChange = (MaxVolume - MinVolume) / 100;
      trackBarVolume.LargeChange = (MaxVolume - MinVolume) / 10;
      trackBarVolume.Value = DefaultVolume;
      trackBarVolume.ValueChanged += delegate {
        SetPlayerVolume();
      };
      trackBarVolume.Scroll += delegate {
        SetVolumeToolTip();
      };
      trackBarVolume.MouseHover += delegate {
        SetVolumeToolTip();
      };

      buttonPlay.Click += delegate {
        if (player == null || !player.Playing)
          Play();
        else if (player.Pausing)
          Resume();
        else
          Pause();
      };

      buttonPlay.Image = imageButtonPlay;
      buttonPlay.ImageAlign = ContentAlignment.MiddleCenter;

      buttonStop.Click += delegate { Stop(); };
      buttonStop.Image = imageButtonStop;
      buttonStop.ImageAlign = ContentAlignment.MiddleCenter;
      buttonStop.Enabled = false;

      buttonPrev.Click += delegate { OnPrevButtonClick(EventArgs.Empty); };
      buttonPrev.Image = imageButtonPrev;
      buttonPrev.ImageAlign = ContentAlignment.MiddleCenter;

      buttonNext.Click += delegate { OnNextButtonClick(EventArgs.Empty); };
      buttonNext.Image = imageButtonNext;
      buttonNext.ImageAlign = ContentAlignment.MiddleCenter;

      SetPrevNextButtonVisibility(showPrevNextButtons);

      checkBoxRepeat.CheckedChanged += delegate {
        if (player != null)
          player.Repeat = checkBoxRepeat.Checked;
      };

      checkBoxAutoScroll.CheckedChanged += delegate {
        waveFormView.SeekBarAutoTracking = checkBoxAutoScroll.Checked;
      };

      SetPlayerVolume();

      timer = new System.Timers.Timer(100);
      timer.Enabled = false;
      timer.AutoReset = true;
      timer.SynchronizingObject = this;
      timer.Elapsed += delegate { SetSeekBarPosition(); };

      labelPosition.Left = checkBoxAutoScroll.Right + 3; // ???: anchor not works
    }

    private void Player_PlayDone(object sender, EventArgs e)
    {
      DisableStopButton();

      OnPlayDone(EventArgs.Empty);
      OnPlayerStatusChanged(EventArgs.Empty);
    }

    protected override void OnLoad(EventArgs e)
    {
      timer.Start();

      base.OnLoad(e);
    }

    private void DisposeInternal(bool disposing)
    {
      if (disposing) {
        if (!disposed) {
          Stop();

          if (timer != null) {
            timer.Stop();
            timer.Dispose();
            timer = null;
          }

          if (player != null) {
            player.PlayDone -= Player_PlayDone;
            player.Close();
            player = null;
          }

          if (stream != null) {
            stream.Close();
            stream = null;
          }

          disposed = true;
        }
      }
    }

    private void SetPrevNextButtonVisibility(bool visible)
    {
      showPrevNextButtons = visible;

      buttonPrev.Visible = showPrevNextButtons;
      buttonNext.Visible = showPrevNextButtons;
    }

    private void SetVolumeToolTip()
    {
      toolTipVolume.SetToolTip(trackBarVolume, GetVolumeInPercent().ToString("P0"));
    }

    /*
     * public methods
     */
    public void Play()
    {
      RejectDisposed();

      if (track == null)
        return;

      if (player == null) {
        player = PcmStreamPlayer.Create();
        player.PlayDone += Player_PlayDone;
      }

      buttonPlay.Image = imageButtonPause;
      buttonStop.Enabled = true;

      try {
        stream = track.GetStream(BgmSourcePath);
      }
      catch (FileNotFoundException) {
        throw;
      }
      catch (ArgumentException ex) {
        DisablePlayButton(ex);
        return;
      }

      try {
        SetPlayerVolume();
      }
      catch (Smdn.Windows.Multimedia.WinMMException ex) {
        player?.Close();
        player = null;
        DisablePlayButton(ex);
        return;
      }

      try {
        player.PlayAsync(stream, stream.Format.GetWaveFormatEx(), checkBoxRepeat.Checked);
      }
      catch (NotSupportedException ex) {
        player?.Close();
        player = null;
        DisablePlayButton(ex);
        return;
      }

      OnPlayerStatusChanged(EventArgs.Empty);

      void DisablePlayButton(Exception exception)
      {
        buttonPlay.Image = imageButtonCanNotPlay;
        buttonStop.Enabled = false;

        toolTipButtonPlay.SetToolTip(buttonPlay, string.Format("{0}: {1}", exception.GetType().FullName, exception.Message));
      }
    }

    public void Stop()
    {
      RejectDisposed();

      if (player == null)
        return;

      DisableStopButton();

      player.Stop();

      OnPlayerStatusChanged(EventArgs.Empty);
    }

    public void Pause()
    {
      RejectDisposed();

      if (player == null)
        return;

      buttonPlay.Image = imageButtonPlay;

      player.Pause();

      OnPlayerStatusChanged(EventArgs.Empty);
    }

    public void Resume()
    {
      RejectDisposed();

      if (player == null)
        return;

      buttonPlay.Image = imageButtonPause;

      player.Resume();

      OnPlayerStatusChanged(EventArgs.Empty);
    }

    /*
     * private methods
     */
    private double GetVolumeInPercent()
    {
      return (trackBarVolume.Value - trackBarVolume.Minimum) / (double)(trackBarVolume.Maximum - trackBarVolume.Minimum);
    }

    private void SetPlayerVolume()
    {
      if (player == null)
        return;

      player.Volume = (int)(100 * GetVolumeInPercent());
    }

    private void DisableStopButton()
    {
      if (InvokeRequired) {
        Invoke((MethodInvoker)DisableStopButton);
        return;
      }

      EnablePlayButton();

      buttonStop.Enabled = false;
    }

    private void EnablePlayButton()
    {
      buttonPlay.Image = imageButtonPlay;
      toolTipButtonPlay.SetToolTip(buttonPlay, null);
    }

    private void SetSeekBarPosition()
    {
      if (player == null)
        return;

      var position = player.Position;

      labelPosition.Text = string.Format("{0}:{1:D2}.{2:D3}", (int)position.TotalMinutes, position.Seconds, position.Milliseconds);

      if (IsPlaying && !IsPausing)
        waveFormView.SeekBarPosition = position;
    }

    private void OnTrackChanged(Track newTrack)
    {
      if (track == newTrack)
        return;

      Stop();

      if (stream != null) {
        stream.Close();
        stream = null;
      }

      track = newTrack;

      waveFormView.Track = track;

      effectEditor.Effect = (track == null) ? null : track.Effect;

      EnablePlayButton();

      OnTrackChanged(EventArgs.Empty);
    }

    protected virtual void OnTrackChanged(EventArgs e)
    {
      var ev = this.TrackChanged;

      if (ev != null)
        ev(this, e);
    }

    protected virtual void OnPlayDone(EventArgs e)
    {
      var ev = this.PlayDone;

      if (ev != null)
        ev(this, e);
    }

    protected virtual void OnPlayerStatusChanged(EventArgs e)
    {
      var ev = this.PlayerStatusChanged;

      if (ev != null)
        ev(this, e);
    }

    protected virtual void OnPrevButtonClick(EventArgs e)
    {
      var ev = this.PrevButtonClick;

      if (ev != null)
        ev(this, e);
    }

    protected virtual void OnNextButtonClick(EventArgs e)
    {
      var ev = this.NextButtonClick;

      if (ev != null)
        ev(this, e);
    }

    private void RejectDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().Name);
    }

    private bool disposed = false;
    private PcmStreamPlayer player;
    private bool showPrevNextButtons = true;
    private Track track = null;
    private string bgmSourcePath = null;
    private EffectAppliedBgmStream stream = null;
    private System.Timers.Timer timer = null;
  }
}
